#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Some extra bits.

:copyright: 2016 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np

from . import utils
from .curve import Curve


class RGBLog(object):
    """
    Attempt at RGB. Incomplete.
    """
    def __init__(self, curves):
        pass
